@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* alias.c: alias substitution */
-/* (C) 2007-2012 magicant */
+/* (C) 2007-2017 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -37,6 +37,7 @@
 #include "option.h"
 #include "parser.h"
 #include "plist.h"
+#include "refcount.h"
 #include "strbuf.h"
 #include "util.h"
 #if YASH_ENABLE_LINEEDIT
@@ -45,14 +46,9 @@
 #endif
 
 
-typedef enum {
-    AF_BLANKEND  = 1 << 0,  /* alias value ends with a blank */
-    AF_GLOBAL    = 1 << 1,  /* is a global alias */
-} aliasflags_T;
-
 typedef struct alias_T {
-    aliasflags_T flags;
-    unsigned refcount;   /* reference count */
+    bool isglobal;
+    refcount_T refcount;
     size_t valuelen;     /* length of `value' */
     wchar_t value[];
 } alias_T;
@@ -66,7 +62,10 @@ typedef struct aliaslist_T {
  * infinite recursive substitution of an alias. When an alias is substituted,
  * the alias and the end index of the substituted string are saved in the list.
  * Substitution of the same alias is not performed before the saved index,
- * thus preventing recursive substitution. */
+ * thus preventing recursive substitution.
+ * `aliaslist_T' is also used to indicate which command words are subject to
+ * substitution after another substitution that ends with a blank.
+ * Alias list items are sorted in the order of `limitindex'. */
 
 static void free_alias(alias_T *alias);
 static inline void vfreealias(kvpair_T kv);
@@ -76,16 +75,17 @@ static void define_alias(
 static bool remove_alias(const wchar_t *name)
     __attribute__((nonnull));
 static void remove_all_aliases(void);
-static aliaslist_T *clone_aliaslist(const aliaslist_T *list)
-    __attribute__((malloc,warn_unused_result));
-static bool contained_in_list(const aliaslist_T *list, const alias_T *alias)
+static bool contained_in_list(
+	const aliaslist_T *list, const alias_T *alias, size_t i)
     __attribute__((pure));
 static void add_to_aliaslist(
 	aliaslist_T **list, alias_T *alias, size_t limitindex)
     __attribute__((nonnull));
-static aliaslist_T *remove_expired_aliases(aliaslist_T *list, size_t index)
-    __attribute__((warn_unused_result));
-static void shift_index(aliaslist_T *list, ptrdiff_t inc);
+static bool remove_expired_aliases(
+	aliaslist_T **list, size_t index, const xwcsbuf_T *buf)
+    __attribute__((nonnull));
+static bool is_after_blank(size_t i, size_t j, const xwcsbuf_T *buf)
+    __attribute__((nonnull));
 static bool is_redir_fd(const wchar_t *s)
     __attribute__((nonnull,pure));
 static bool print_alias(const wchar_t *name, const alias_T *alias, bool prefix);
@@ -112,11 +112,9 @@ inline bool is_alias_name_char(wchar_t c)
  * frees it. This function does nothing if `alias' is a null pointer. */
 void free_alias(alias_T *alias)
 {
-    if (alias != NULL) {
-	alias->refcount--;
-	if (alias->refcount == 0)
+    if (alias != NULL)
+	if (refcount_decrement(&alias->refcount))
 	    free(alias);
-    }
 }
 
 /* Applies `free_alias' to the value of key-value pair `kv'. */
@@ -137,15 +135,11 @@ void define_alias(
     size_t namelen = equal - nameandvalue;
     size_t valuelen = wcslen(equal + 1);
     alias_T *alias = xmallocs(sizeof *alias,
-	    namelen + valuelen + 2, sizeof *alias->value);
+	    add(add(namelen, valuelen), 2), sizeof *alias->value);
 
-    alias->flags = 0;
+    alias->isglobal = global;
     alias->refcount = 1;
     alias->valuelen = valuelen;
-    if (global)
-	alias->flags |= AF_GLOBAL;
-    if (iswblank(equal[valuelen]))  // `(equal + 1)[valuelen - 1]'
-	alias->flags |= AF_BLANKEND;
     wmemcpy(alias->value, equal + 1, valuelen);
     alias->value[valuelen] = L'\0';
     wmemcpy(alias->value + valuelen + 1, nameandvalue, namelen);
@@ -186,39 +180,23 @@ const wchar_t *get_alias_value(const wchar_t *aliasname)
 	return NULL;
 }
 
-/* Creates a copy of the specified alias list. */
-aliaslist_T *clone_aliaslist(const aliaslist_T *list)
-{
-    aliaslist_T  *result = NULL;
-    aliaslist_T **lastp  = &result;
-
-    while (list != NULL) {
-	aliaslist_T *copy = xmalloc(sizeof *copy);
-
-	copy->next = NULL;
-	copy->alias = list->alias;
-	copy->alias->refcount++;
-	copy->limitindex = list->limitindex;
-
-	*lastp = copy;
-	lastp = &copy->next;
-	list = list->next;
-    }
-    return result;
-}
-
 /* Frees the specified alias list and its contents. */
 void destroy_aliaslist(aliaslist_T *list)
 {
-    list = remove_expired_aliases(list, SIZE_MAX);
-    assert(list == NULL);
+    while (list != NULL) {
+	aliaslist_T *next = list->next;
+	free_alias(list->alias);
+	free(list);
+	list = next;
+    }
 }
 
-/* Checks if the specified alias list contains the specified alias. */
-bool contained_in_list(const aliaslist_T *list, const alias_T *alias)
+/* Checks if the specified alias list contains the specified alias.
+ * List items whose limit index is less than `i' are ignored. */
+bool contained_in_list(const aliaslist_T *list, const alias_T *alias, size_t i)
 {
     while (list != NULL) {
-	if (list->alias == alias)
+	if (list->limitindex >= i && list->alias == alias)
 	    return true;
 	list = list->next;
     }
@@ -228,117 +206,144 @@ bool contained_in_list(const aliaslist_T *list, const alias_T *alias)
 /* Adds an alias to an alias list with the specified limit index. */
 void add_to_aliaslist(aliaslist_T **list, alias_T *alias, size_t limitindex)
 {
-    aliaslist_T *oldhead = *list;
-    aliaslist_T *newelem = xmalloc(sizeof *newelem);
+    /* Find where to insert the new item. Remember, the list is sorted in the
+     * order of `limitindex'. */
+    while (*list != NULL && (*list)->limitindex < limitindex)
+	list = &(*list)->next;
 
-    assert(oldhead == NULL || limitindex <= oldhead->limitindex);
-    newelem->next = oldhead;
+    aliaslist_T *newelem = xmalloc(sizeof *newelem);
+    newelem->next = *list;
     newelem->alias = alias;
-    newelem->alias->refcount++;
+    refcount_increment(&newelem->alias->refcount);
     newelem->limitindex = limitindex;
     *list = newelem;
 }
 
-/* Removes items whose `limitindex' is less than or equal to `index'. */
-aliaslist_T *remove_expired_aliases(aliaslist_T *list, size_t index)
+/* Removes items that are no longer significant. An item is significant if (1)
+ * its limit index is larger than `index', or (2) it is a non-global alias and
+ * all the characters are blank between the indexes `limitindex-1' and `index'.
+ * Returns true iff a significant item of type (2) is left. */
+/* If this function returns true, the `index' is just after the result of alias
+ * substitution that ends with a blank, in which case the next word should be
+ * checked for another substitution. */
+bool remove_expired_aliases(
+	aliaslist_T **list, size_t index, const xwcsbuf_T *buf)
 {
+    aliaslist_T *item = *list;
+    bool afterblank = false;
+
     /* List items are ordered by index; we don't have to check all the items. */
-    while (list != NULL && list->limitindex <= index) {
-	aliaslist_T *next = list->next;
-	free_alias(list->alias);
-	free(list);
-	list = next;
+    while (item != NULL && item->limitindex <= index) {
+	if (!item->alias->isglobal
+		&& is_after_blank(item->limitindex, index, buf)) {
+	    afterblank = true;
+	    break;
+	}
+
+	aliaslist_T *next = item->next;
+	free_alias(item->alias);
+	free(item);
+	item = next;
     }
-    return list;
+    *list = item;
+    return afterblank;
 }
 
-/* Increases the index of each item in the specified list by `inc'. */
-void shift_index(aliaslist_T *list, ptrdiff_t inc)
+/* Tests if the character just before `i` in `buf' is a blank and all the
+ * characters between `i' and `j' are blanks. */
+bool is_after_blank(size_t i, size_t j, const xwcsbuf_T *buf)
+{
+    assert(i <= j);
+    assert(j <= buf->length);
+
+    if (i == 0)
+	return false;
+    for (i--; i < j; i++)
+	if (!iswblank(buf->contents[i]))
+	    return false;
+    return true;
+}
+
+/* Increases the limit index by `inc' for each item whose index is larger
+ * than `i'. */
+void shift_aliaslist_index(aliaslist_T *list, size_t i, ptrdiff_t inc)
 {
     while (list != NULL) {
-	list->limitindex += inc;
+	if (list->limitindex > i) {
+	    assert(inc >= 0 || (size_t) -inc <= list->limitindex);
+	    list->limitindex += inc;
+	}
 	list = list->next;
     }
 }
 
-/* Performs alias substitution at index `i' in buffer `buf'.
- * If AF_NONGLOBAL is not in `flags', only global aliases are substituted.
- * If AF_NORECUR is not in `flags', substitution is repeated until there is
- * no more alias applicable.
- * Returns true iff any alias was substituted. */
+/* Performs alias substitution on the word starting at index `i' in buffer
+ * `buf'. */
 bool substitute_alias(xwcsbuf_T *restrict buf, size_t i,
 	aliaslist_T **restrict list, substaliasflags_T flags)
+{
+    size_t j = i;
+    while (is_alias_name_char(buf->contents[j]))
+	j++;
+    return substitute_alias_range(buf, i, j, list, flags);
+}
+
+/* Performs alias substitution on the word in the index range `[i, j)` in buffer
+ * `buf'.
+ * If AF_NONGLOBAL is not in `flags' and `i' is not after another substitution
+ * that ends with a blank, only global aliases are substituted.
+ * The substitution is not recursive: the resultant string may be another alias
+ * that should be substituted by calling `substitute_alias' again.
+ * Returns true iff any alias was substituted. */
+bool substitute_alias_range(xwcsbuf_T *restrict buf, size_t i, size_t j,
+	aliaslist_T **restrict list, substaliasflags_T flags)
 {
     if (aliases.count == 0)
 	return false;
+
+    if (remove_expired_aliases(list, i, buf))
+	flags |= AF_NONGLOBAL;
+
     if (!(flags & AF_NONGLOBAL) && posixly_correct)
 	return false;
 
-    bool subst = false;
-
-substitute_alias:
-    *list = remove_expired_aliases(*list, i);
-
-    /* count the length of the alias name */
-    size_t j = i;
-    while (is_alias_name_char(buf->contents[j]))
-	j++;
-    /* `i' is the starting index of the alias name and `j' is the ending index*/
-
     /* check if there is an alias name */
-    if (i < j && is_token_delimiter_char(buf->contents[j])
-	    && !is_redir_fd(buf->contents + i)) {
+    if (i >= j)
+	return false;
+    if (flags & AF_NOEOF)
+	if (j == buf->length)
+	    return false;
+    if (!is_token_delimiter_char(buf->contents[j]))
+	return false;
+    if (is_redir_fd(buf->contents + i))
+	return false;
 
-	alias_T *alias;
-	wchar_t savechar;
+    alias_T *alias;
 
-	/* get alias definition */
-	savechar = buf->contents[j];
-	buf->contents[j] = L'\0';
-	alias = ht_get(&aliases, buf->contents + i).value;
-	buf->contents[j] = savechar;
+    /* get alias definition */
+    wchar_t savechar = buf->contents[j];
+    buf->contents[j] = L'\0';
+    alias = ht_get(&aliases, buf->contents + i).value;
+    buf->contents[j] = savechar;
 
-	/* check if we should do substitution */
-	if (alias != NULL
-		&& ((flags & AF_NONGLOBAL) | (alias->flags & AF_GLOBAL))
-		&& !contained_in_list(*list, alias)) {
+    /* check if we should do substitution */
+    if (alias == NULL)
+	return false;
+    if (!(flags & AF_NONGLOBAL) && !alias->isglobal)
+	return false;
+    if (contained_in_list(*list, alias, i))
+	return false;
 
-	    /* do substitution */
-	    wb_replace_force(buf, i, j - i, alias->value, alias->valuelen);
-	    shift_index(*list,
-		    (ptrdiff_t) alias->valuelen - (ptrdiff_t) (j - i));
-	    subst = true;
+    /* do substitution */
+    wb_replace_force(buf, i, j - i, alias->value, alias->valuelen);
+    shift_aliaslist_index(
+	    *list, i, (ptrdiff_t) alias->valuelen - (ptrdiff_t) (j - i));
 
-	    /* substitute the following word if AF_BLANKEND is set */
-	    /* see note below */
-	    if ((alias->flags & AF_BLANKEND) && !(alias->flags & AF_GLOBAL)) {
-		size_t ii = i + alias->valuelen;
-		aliaslist_T *savelist = clone_aliaslist(*list);
-		while (iswblank(buf->contents[ii]))
-		    ii++;
-		substitute_alias(buf, ii, &savelist, flags);
-		destroy_aliaslist(savelist);
-	    }
+    /* add the alias to the list to track recursion */
+    add_to_aliaslist(list, alias, i + alias->valuelen);
 
-	    /* add the alias to the list to track recursion */
-	    add_to_aliaslist(list, alias, i + alias->valuelen);
-
-	    /* recursively substitute alias */
-	    if (!(flags & AF_NORECUR)) {
-		while (iswblank(buf->contents[i]))
-		    i++;
-		goto substitute_alias;
-	    }
-	}
-    }
-    return subst;
+    return true;
 }
-/* When the value of the alias ends with a blank, we substitute the following
- * word if and only if the alias is not global. This is required to prevent
- * infinite substitution that would happen in some cases such as:
- *    alias -g a='a a '
- *    echo a
- */
 
 /* Returns true iff the specified string starts with any number of digits
  * followed by L'<' or L'>'. */
@@ -363,7 +368,7 @@ bool print_alias(const wchar_t *name, const alias_T *alias, bool prefix)
 
     if (!prefix)
 	format = "%ls=%ls\n";
-    else if (alias->flags & AF_GLOBAL)
+    else if (alias->isglobal)
 	if (name[0] == L'-')
 	    format = "alias -g -- %ls=%ls\n";
 	else
@@ -387,7 +392,7 @@ bool print_alias_if_defined(const wchar_t *aliasname, bool user_friendly)
 {
     const alias_T *alias = ht_get(&aliases, aliasname).value;
 
-    if (alias == NULL || (alias->flags & AF_GLOBAL))
+    if (alias == NULL || alias->isglobal)
 	return false;
 
     if (!user_friendly)
@@ -415,8 +420,7 @@ void generate_alias_candidates(const le_compopt_T *compopt)
 
     while ((kv = ht_next(&aliases, &i)).key != NULL) {
 	const alias_T *alias = kv.value;
-	le_candgentype_T type =
-	    (alias->flags & AF_GLOBAL) ? CGT_GALIAS : CGT_NALIAS;
+	le_candgentype_T type = alias->isglobal ? CGT_GALIAS : CGT_NALIAS;
 
 	if (compopt->type & type)
 	    if (le_wmatch_comppatterns(compopt, kv.key))
