@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* alias.h: alias substitution */
-/* (C) 2007-2012 magicant */
+/* (C) 2007-2017 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -20,7 +20,7 @@
 #define YASH_ALIAS_H
 
 #include <stddef.h>
-#include "util.h"
+#include "xgetopt.h"
 
 
 struct xwcsbuf_T;
@@ -28,7 +28,7 @@ struct aliaslist_T;
 
 typedef enum {
     AF_NONGLOBAL = 1 << 0,
-    AF_NORECUR   = 1 << 1,
+    AF_NOEOF     = 1 << 1,
 } substaliasflags_T;
 
 extern void init_alias(void);
@@ -37,22 +37,32 @@ extern _Bool is_alias_name_char(wchar_t c)
 extern const wchar_t *get_alias_value(const wchar_t *aliasname)
     __attribute__((nonnull,pure));
 extern void destroy_aliaslist(struct aliaslist_T *list);
+extern void shift_aliaslist_index(
+	struct aliaslist_T *list, size_t i, ptrdiff_t inc);
 extern _Bool substitute_alias(
 	struct xwcsbuf_T *restrict buf, size_t i,
 	struct aliaslist_T **restrict list, substaliasflags_T flags)
     __attribute__((nonnull));
+extern _Bool substitute_alias_range(
+	struct xwcsbuf_T *restrict buf, size_t i, size_t j,
+	struct aliaslist_T **restrict list, substaliasflags_T flags)
+    __attribute__((nonnull));
 extern _Bool print_alias_if_defined(
 	const wchar_t *aliasname, _Bool user_friendly)
     __attribute__((nonnull));
 
 extern int alias_builtin(int argc, void **argv)
     __attribute__((nonnull));
+#if YASH_ENABLE_HELP
 extern const char alias_help[], alias_syntax[];
+#endif
 extern const struct xgetopt_T alias_options[];
 
 extern int unalias_builtin(int argc, void **argv)
     __attribute__((nonnull));
+#if YASH_ENABLE_HELP
 extern const char unalias_help[], unalias_syntax[];
+#endif
 extern const struct xgetopt_T unalias_options[];
 
 
