@@ -121,8 +121,7 @@ typedef enum exception_T {
 /* state of currently executed loop */
 typedef struct execstate_T {
     unsigned loopnest;      /* level of nested loops */
-    unsigned breakloopnest; /* target of break/continue */
-    exception_T exception;  /* exceptional jump to be done */
+    unsigned breakloopnest; /* target level of "break" */
     bool noreturn;          /* true when the "return" built-in is not allowed */
     bool iterating;         /* true when iterative execution is ongoing */
 } execstate_T;
@@ -160,8 +159,6 @@ static void search_command(
     __attribute__((nonnull));
 static inline bool is_special_builtin(const char *cmdname)
     __attribute__((nonnull,pure));
-static inline bool assignment_is_temporary(enum cmdtype_T type)
-    __attribute__((const));
 static bool command_not_found_handler(void *const *argv)
     __attribute__((nonnull));
 static void exec_nonsimple_command(command_T *c, bool finally_exit)
@@ -203,7 +200,8 @@ static bool suppresserrexit = false;
 
 /* state of currently executed loop */
 static execstate_T execstate;
-/* Note that `execstate' is not reset when a subshell forks. */
+/* exceptional jump to be done (other than "break") */
+static exception_T exception;
 
 /* This flag is set when a special built-in is executed as such. */
 bool special_builtin_executed;
@@ -223,29 +221,26 @@ static xwcsbuf_T xtrace_buffer = { .contents = NULL };
 
 
 /* Resets `execstate' to the initial state. */
-void reset_execstate(void)
+void reset_execstate(bool reset_iteration)
 {
-    execstate = (struct execstate_T) {
-	.loopnest = 0,
-	.breakloopnest = UINT_MAX,
-	.exception = E_NONE,
-	.noreturn = false,
-	.iterating = false,
-    };
+    execstate.loopnest = 0;
+    execstate.breakloopnest = 0;
+    execstate.noreturn = false;
+    if (reset_iteration)
+	execstate.iterating = false;
 }
 
 /* Saves the current `execstate' and returns it.
- * `execstate' is re-initialized. */
-struct execstate_T *save_execstate(void)
+ * You typically call `reset_execstate' after calling this function. */
+execstate_T *save_execstate(void)
 {
-    struct execstate_T *save = xmalloc(sizeof execstate);
+    execstate_T *save = xmalloc(sizeof execstate);
     *save = execstate;
-    reset_execstate();
     return save;
 }
 
 /* Restores `execstate' to `save' and frees `save'. */
-void restore_execstate(struct execstate_T *save)
+void restore_execstate(execstate_T *save)
 {
     execstate = *save;
     free(save);
@@ -260,15 +255,15 @@ void disable_return(void)
 /* If we're returning, clear the flag. */
 void cancel_return(void)
 {
-    if (execstate.exception == E_RETURN)
-	execstate.exception = E_NONE;
+    if (exception == E_RETURN)
+	exception = E_NONE;
 }
 
 /* Returns true iff we're breaking/continuing/returning now. */
 bool need_break(void)
 {
     return execstate.breakloopnest < execstate.loopnest
-	|| execstate.exception != E_NONE
+	|| exception != E_NONE
 	|| is_interrupted();
 }
 
@@ -399,7 +394,7 @@ void exec_for(const command_T *c, bool finally_exit)
 {
     assert(c->c_type == CT_FOR);
     execstate.loopnest++;
-    execstate.breakloopnest = UINT_MAX;
+    execstate.breakloopnest = execstate.loopnest;
 
     int count;
     void **words;
@@ -424,10 +419,10 @@ void exec_for(const command_T *c, bool finally_exit)
 #define CHECK_LOOP                                      \
     if (execstate.breakloopnest < execstate.loopnest) { \
 	goto done;                                      \
-    } else if (execstate.exception == E_CONTINUE) {     \
-	execstate.exception = E_NONE;                   \
+    } else if (exception == E_CONTINUE) {               \
+	exception = E_NONE;                             \
 	continue;                                       \
-    } else if (execstate.exception != E_NONE) {         \
+    } else if (exception != E_NONE) {                   \
 	goto done;                                      \
     } else if (is_interrupted()) {                      \
 	goto done;                                      \
@@ -467,7 +462,7 @@ void exec_while(const command_T *c, bool finally_exit)
 {
     assert(c->c_type == CT_WHILE);
     execstate.loopnest++;
-    execstate.breakloopnest = UINT_MAX;
+    execstate.breakloopnest = execstate.loopnest;
 
     int status = Exit_SUCCESS;
     for (;;) {
@@ -852,7 +847,7 @@ pid_t exec_process(
      * whether we have to open a temporary environment. */
     search_command(argv0, argv[0], &cmdinfo, SCT_BUILTIN | SCT_FUNCTION);
     special_builtin_executed = (cmdinfo.type == CT_SPECIALBUILTIN);
-    temp = c->c_assigns != NULL && assignment_is_temporary(cmdinfo.type);
+    temp = c->c_assigns != NULL && !special_builtin_executed;
     if (temp)
 	open_new_environment(true);
 
@@ -861,8 +856,7 @@ pid_t exec_process(
 	/* On assignment error, the command is not executed. */
 	print_xtrace(NULL);
 	laststatus = Exit_ASSGNERR;
-	if (posixly_correct && !is_interactive_now &&
-		cmdinfo.type == CT_SPECIALBUILTIN)
+	if (!is_interactive_now)
 	    finally_exit = true;
 	goto done3;
     }
@@ -998,7 +992,7 @@ pid_t fork_and_reset(pid_t pgid, bool fg, sigtype_T sigtype)
 		put_foreground(pgid);
 	}
 	if (sigtype & t_quitint)
-	    if (posixly_correct || !save_doing_job_control_now)
+	    if (!save_doing_job_control_now)
 		ignore_sigquit_and_sigint();
 	if (sigtype & t_tstp)
 	    if (save_doing_job_control_now)
@@ -1011,7 +1005,7 @@ pid_t fork_and_reset(pid_t pgid, bool fg, sigtype_T sigtype)
 #if YASH_ENABLE_HISTORY
 	    close_history_file();
 #endif
-	    reset_execstate();
+	    reset_execstate(true);
 	}
 	restore_signals(sigtype & t_leave);  /* signal mask is restored here */
 	clear_shellfds(sigtype & t_leave);
@@ -1109,22 +1103,6 @@ bool is_special_builtin(const char *cmdname)
     return bi != NULL && bi->type == BI_SPECIAL;
 }
 
-/* Determines whether the assignments should be temporary according to `type'.*/
-bool assignment_is_temporary(enum cmdtype_T type)
-{
-    switch (type) {
-	case CT_SPECIALBUILTIN:
-	case CT_FUNCTION:
-	    return false;
-	case CT_NONE:
-	case CT_SEMISPECIALBUILTIN:
-	case CT_REGULARBUILTIN:
-	case CT_EXTERNALPROGRAM:
-	    return true;
-    }
-    assert(false);
-}
-
 /* Executes $COMMAND_NOT_FOUND_HANDLER if any.
  * `argv' is set to the positional parameters of the environment in which the
  * handler is executed.
@@ -1370,8 +1348,8 @@ void exec_fall_back_on_sh(
 void exec_function_body(
 	command_T *body, void *const *args, bool finally_exit, bool complete)
 {
-    bool save_noreturn = execstate.noreturn;
-    execstate.noreturn = false;
+    execstate_T *saveexecstate = save_execstate();
+    reset_execstate(false);
 
     open_new_environment(false);
     set_positional_parameters(args);
@@ -1382,10 +1360,10 @@ void exec_function_body(
     (void) complete;
 #endif
     exec_commands(body, finally_exit ? E_SELF : E_NORMAL);
-    cancel_return();
     close_current_environment();
 
-    execstate.noreturn = save_noreturn;
+    cancel_return();
+    restore_execstate(saveexecstate);
 }
 
 /* Calls `execve' until it doesn't return EINTR. */
@@ -1494,14 +1472,14 @@ int exec_iteration(void *const *commands, const char *codename)
 	exec_wcs(*command, codename, false);
 	commandstatus = laststatus;
 	laststatus = savelaststatus;
-	switch (execstate.exception) {
+	switch (exception) {
 	    case E_BREAK_ITERATION:
-		execstate.exception = E_NONE;
+		exception = E_NONE;
 		/* falls thru! */
 	    default:
 		goto done;
 	    case E_CONTINUE_ITERATION:
-		execstate.exception = E_NONE;
+		exception = E_NONE;
 		/* falls thru! */
 	    case E_NONE:
 		break;
@@ -1541,7 +1519,15 @@ int exec_variable_as_commands(const wchar_t *varname, const char *codename)
     /* copy the array values in case they are unset during execution */
     save_get_variable_values(&gv);
 
+    /* prevent "break" and "continue" */
+    execstate_T *saveexecstate = save_execstate();
+    // reset_execstate(false); /* don't reset `execstate.noreturn' */
+    execstate.loopnest = 0;
+
     int result = exec_iteration(gv.values, codename);
+
+    restore_execstate(saveexecstate);
+
     plfree(gv.values, free);
     return result;
 }
@@ -1586,9 +1572,10 @@ bool autoload_completion_function_file(
 	return false;
     }
 
-    struct execstate_T *saveexecstate = save_execstate();
+    execstate_T *saveexecstate = save_execstate();
     int savelaststatus = laststatus;
     bool saveposix = posixly_correct;
+    reset_execstate(true);
     posixly_correct = false;
     open_new_environment(false);
     set_positional_parameters((void *[]) { (void *) cmdname, NULL });
@@ -1600,6 +1587,7 @@ bool autoload_completion_function_file(
     close_current_environment();
     posixly_correct = saveposix;
     laststatus = savelaststatus;
+    cancel_return();
     restore_execstate(saveexecstate);
     remove_shellfd(fd);
     xclose(fd);
@@ -1618,9 +1606,10 @@ bool call_completion_function(const wchar_t *funcname)
 	return false;
     }
 
-    struct execstate_T *saveexecstate = save_execstate();
+    execstate_T *saveexecstate = save_execstate();
     int savelaststatus = laststatus;
     bool saveposix = posixly_correct;
+    reset_execstate(true);
     posixly_correct = false;
 
     le_compdebug("executing completion function \"%ls\"", funcname);
@@ -1632,6 +1621,7 @@ bool call_completion_function(const wchar_t *funcname)
 
     posixly_correct = saveposix;
     laststatus = savelaststatus;
+    cancel_return();
     restore_execstate(saveexecstate);
 
     return true;
@@ -1690,12 +1680,12 @@ int return_builtin(int argc, void **argv)
 		return print_builtin_help(ARGV(0));
 #endif
 	    default:
-		return special_builtin_syntax_error(Exit_ERROR);
+		return special_builtin_error(Exit_ERROR);
 	}
     }
 
     if (!validate_operand_count(argc - xoptind, 0, 1))
-	return special_builtin_syntax_error(Exit_ERROR);
+	return special_builtin_error(Exit_ERROR);
 
     int status;
     const wchar_t *statusstr = ARGV(xoptind);
@@ -1703,7 +1693,7 @@ int return_builtin(int argc, void **argv)
 	if (!xwcstoi(statusstr, 10, &status) || status < 0) {
 	    xerror(0, Ngt("`%ls' is not a valid integer"), statusstr);
 	    status = Exit_ERROR;
-	    special_builtin_syntax_error(status);
+	    special_builtin_error(status);
 	    /* return anyway */
 	}
     } else {
@@ -1714,7 +1704,7 @@ int return_builtin(int argc, void **argv)
 	    xerror(0, Ngt("cannot be used in the interactive mode"));
 	    return Exit_FAILURE;
 	}
-	execstate.exception = E_RETURN;
+	exception = E_RETURN;
     }
     return status;
 }
@@ -1746,12 +1736,12 @@ int break_builtin(int argc, void **argv)
 		return print_builtin_help(ARGV(0));
 #endif
 	    default:
-		return special_builtin_syntax_error(Exit_ERROR);
+		return special_builtin_error(Exit_ERROR);
 	}
     }
 
     if (!validate_operand_count(argc - xoptind, 0, iter ? 0 : 1))
-	return special_builtin_syntax_error(Exit_ERROR);
+	return special_builtin_error(Exit_ERROR);
 
     if (iter) {
 	/* break/continue iteration */
@@ -1760,10 +1750,10 @@ int break_builtin(int argc, void **argv)
 	    return Exit_ERROR;
 	}
 	if (wcscmp(ARGV(0), L"break") == 0) {
-	    execstate.exception = E_BREAK_ITERATION;
+	    exception = E_BREAK_ITERATION;
 	} else {
 	    assert(wcscmp(ARGV(0), L"continue") == 0);
-	    execstate.exception = E_CONTINUE_ITERATION;
+	    exception = E_CONTINUE_ITERATION;
 	}
 	return laststatus;
     } else {
@@ -1775,10 +1765,10 @@ int break_builtin(int argc, void **argv)
 	    unsigned long countl;
 	    if (!xwcstoul(countstr, 0, &countl)) {
 		xerror(0, Ngt("`%ls' is not a valid integer"), countstr);
-		return special_builtin_syntax_error(Exit_ERROR);
+		return special_builtin_error(Exit_ERROR);
 	    } else if (countl == 0) {
 		xerror(0, Ngt("%u is not a positive integer"), 0u);
-		return special_builtin_syntax_error(Exit_ERROR);
+		return special_builtin_error(Exit_ERROR);
 	    } else if (countl > UINT_MAX) {
 		count = UINT_MAX;
 	    } else {
@@ -1788,7 +1778,7 @@ int break_builtin(int argc, void **argv)
 	assert(count > 0);
 	if (execstate.loopnest == 0) {
 	    xerror(0, Ngt("not in a loop"));
-	    return special_builtin_syntax_error(Exit_ERROR);
+	    return special_builtin_error(Exit_ERROR);
 	}
 	if (count > execstate.loopnest)
 	    count = execstate.loopnest;
@@ -1797,7 +1787,7 @@ int break_builtin(int argc, void **argv)
 	} else {
 	    assert(wcscmp(ARGV(0), L"continue") == 0);
 	    execstate.breakloopnest = execstate.loopnest - count + 1;
-	    execstate.exception = E_CONTINUE;
+	    exception = E_CONTINUE;
 	}
 	return Exit_SUCCESS;
     }
@@ -1841,7 +1831,7 @@ int eval_builtin(int argc __attribute__((unused)), void **argv)
 		return print_builtin_help(ARGV(0));
 #endif
 	    default:
-		return special_builtin_syntax_error(Exit_ERROR);
+		return special_builtin_error(Exit_ERROR);
 	}
     }
 
@@ -1896,17 +1886,17 @@ int dot_builtin(int argc, void **argv)
 		return print_builtin_help(ARGV(0));
 #endif
 	    default:
-		return special_builtin_syntax_error(Exit_ERROR);
+		return special_builtin_error(Exit_ERROR);
 	}
     }
 
     const wchar_t *filename = ARGV(xoptind++);
     if (filename == NULL)
-	return special_builtin_syntax_error(insufficient_operands_error(1));
+	return special_builtin_error(insufficient_operands_error(1));
 
     bool has_args = xoptind < argc;
     if (has_args && posixly_correct)
-	return special_builtin_syntax_error(too_many_operands_error(1));
+	return special_builtin_error(too_many_operands_error(1));
 
     char *mbsfilename = malloc_wcstombs(filename);
     if (mbsfilename == NULL) {
@@ -1950,8 +1940,11 @@ int dot_builtin(int argc, void **argv)
 	goto error;
     }
 
+    execstate_T *saveexecstate = save_execstate();
+    reset_execstate(false);
     exec_input(fd, mbsfilename, enable_alias ? XIO_SUBST_ALIAS : 0);
     cancel_return();
+    restore_execstate(saveexecstate);
     remove_shellfd(fd);
     xclose(fd);
     free(mbsfilename);
@@ -2010,7 +2003,7 @@ int exec_builtin(int argc, void **argv)
 		return print_builtin_help(ARGV(0));
 #endif
 	    default:
-		return special_builtin_syntax_error(Exit_ERROR);
+		return special_builtin_error(Exit_ERROR);
 	}
     }
 
@@ -2401,12 +2394,12 @@ int times_builtin(int argc __attribute__((unused)), void **argv)
 		return print_builtin_help(ARGV(0));
 #endif
 	    default:
-		return special_builtin_syntax_error(Exit_ERROR);
+		return special_builtin_error(Exit_ERROR);
 	}
     }
 
     if (xoptind < argc)
-	return special_builtin_syntax_error(too_many_operands_error(0));
+	return special_builtin_error(too_many_operands_error(0));
 
     double clock;
     struct tms tms;
@@ -2423,7 +2416,7 @@ int times_builtin(int argc __attribute__((unused)), void **argv)
     clock = sysconf(_SC_CLK_TCK);
     if (times(&tms) == (clock_t) -1) {
 	xerror(errno, Ngt("cannot get the time data"));
-	return Exit_FAILURE;
+	return special_builtin_error(Exit_FAILURE);
     }
     format_time(tms.tms_utime, sum, sus);
     format_time(tms.tms_stime, ssm, sss);
@@ -2433,7 +2426,8 @@ int times_builtin(int argc __attribute__((unused)), void **argv)
 
     xprintf("%jdm%fs %jdm%fs\n%jdm%fs %jdm%fs\n",
 	    sum, sus, ssm, sss, cum, cus, csm, css);
-    return (yash_error_message_count == 0) ? Exit_SUCCESS : Exit_FAILURE;
+    return (yash_error_message_count == 0) ?
+	    Exit_SUCCESS : special_builtin_error(Exit_FAILURE);
 }
 
 #if YASH_ENABLE_HELP
