@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* exec.h: command execution */
-/* (C) 2007-2012 magicant */
+/* (C) 2007-2016 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -21,7 +21,7 @@
 
 #include <stddef.h>
 #include <sys/types.h>
-#include "util.h"
+#include "xgetopt.h"
 
 
 /* options for `fork_and_reset' */
@@ -44,9 +44,10 @@ typedef enum sigtype_T {
 extern int laststatus, savelaststatus;
 extern pid_t lastasyncpid;
 extern _Bool special_builtin_executed;
+extern _Bool is_executing_auxiliary;
 
 struct execstate_T;
-extern void reset_execstate(void);
+extern void reset_execstate(_Bool reset_iteration);
 extern struct execstate_T *save_execstate(void)
     __attribute__((malloc,warn_unused_result));
 extern void restore_execstate(struct execstate_T *save)
@@ -66,47 +67,68 @@ extern wchar_t *exec_command_substitution(const struct embedcmd_T *cmdsub)
 extern int exec_variable_as_commands(
 	const wchar_t *varname, const char *codename)
     __attribute__((nonnull));
+extern int exec_variable_as_auxiliary(
+	const wchar_t *varname, const char *codename)
+    __attribute__((nonnull));
+#define exec_variable_as_auxiliary_(varname) \
+	exec_variable_as_auxiliary(L varname, "$" varname)
 
+#if YASH_ENABLE_LINEEDIT
 extern _Bool autoload_completion_function_file(
 	const wchar_t *filename, const wchar_t *cmdname)
     __attribute__((nonnull(1)));
 extern _Bool call_completion_function(const wchar_t *funcname)
     __attribute__((nonnull));
+#endif
 
 extern const struct xgetopt_T iter_options[];
 
 extern int return_builtin(int argc, void **argv)
     __attribute__((nonnull));
+#if YASH_ENABLE_HELP
 extern const char return_help[], return_syntax[];
+#endif
 extern const struct xgetopt_T return_options[];
 
 extern int break_builtin(int argc, void **argv)
     __attribute__((nonnull));
+#if YASH_ENABLE_HELP
 extern const char break_help[], break_syntax[], continue_help[],
 	continue_syntax[];
+#endif
 
 extern int eval_builtin(int argc, void **argv)
     __attribute__((nonnull));
+#if YASH_ENABLE_HELP
 extern const char eval_help[], eval_syntax[];
+#endif
 
 extern int dot_builtin(int argc, void **argv)
     __attribute__((nonnull));
+#if YASH_ENABLE_HELP
 extern const char dot_help[], dot_syntax[];
+#endif
 extern const struct xgetopt_T dot_options[];
 
 extern int exec_builtin(int argc, void **argv)
     __attribute__((nonnull));
+#if YASH_ENABLE_HELP
 extern const char exec_help[], exec_syntax[];
+#endif
 extern const struct xgetopt_T exec_options[];
 
 extern int command_builtin(int argc, void **argv)
     __attribute__((nonnull));
+#if YASH_ENABLE_HELP
 extern const char command_help[], command_syntax[], type_help[], type_syntax[];
+#endif
 extern const struct xgetopt_T command_options[];
 
 extern int times_builtin(int argc, void **argv)
     __attribute__((nonnull));
+#if YASH_ENABLE_HELP
 extern const char times_help[], times_syntax[];
+#endif
 
 
 #endif /* YASH_EXEC_H */
