@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* expand.c: word expansion */
-/* (C) 2007-2016 magicant */
+/* (C) 2007-2017 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -609,28 +609,23 @@ wchar_t *expand_tilde(const wchar_t **ss, bool hasnextwordunit, tildetype_T tt)
 	return NULL;
     }
     if (!posixly_correct) {
-	size_t INIT(index);
-	if (username[0] == L'+') {
-	    if (username[1] == L'\0') {
-		home = getvar(L VAR_PWD);
-		goto finish;
+	if (username[0] == L'+' && username[1] == L'\0') {
+	    home = getvar(L VAR_PWD);
+	    goto finish;
+	}
+	if (username[0] == L'-' && username[1] == L'\0') {
+	    home = getvar(L VAR_OLDPWD);
+	    goto finish;
+	}
 #if YASH_ENABLE_DIRSTACK
-	    } else if (parse_dirstack_index(username, &index, &home, false)
+	if (username[0] == L'+' || username[0] == L'-') {
+	    size_t index;
+	    if (parse_dirstack_index(username, &index, &home, false)
 		    && index != SIZE_MAX) {
 		goto finish;
-#endif
-	    }
-	} else if (username[0] == L'-') {
-	    if (username[1] == L'\0') {
-		home = getvar(L VAR_OLDPWD);
-		goto finish;
-#if YASH_ENABLE_DIRSTACK
-	    } else if (parse_dirstack_index(username, &index, &home, false)
-		    && index != SIZE_MAX) {
-		goto finish;
-#endif
 	    }
 	}
+#endif
     }
     home = get_home_directory(username, false);
 finish:
@@ -899,17 +894,13 @@ subst:
     }
 
     /* concatenate the elements of `values' */
-    if (concat)
+    if (concat && indq)
 	values = concatenate_values_into_array(values, false);
 
     /* PT_NUMBER */
     if (p->pe_type & PT_NUMBER)
 	subst_length_each(values);
 
-    /* concatenate the elements of `values' */
-    if (!indq)
-	values = concatenate_values_into_array(values, false);
-
     /* backslash escape */
     for (size_t i = 0; values[i] != NULL; i++)
 	values[i] = escapefree(values[i], indq ? NULL : CHARS_ESCAPED);
@@ -1215,7 +1206,7 @@ start:
     pl_init(&splitpoints);
     pl_add(&splitpoints, c);
     nest = 0;
-    while ((c = escaped_wcspbrk(c, L"{,}")) != L'\0') {
+    while ((c = escaped_wcspbrk(c, L"{,}")) != NULL) {
 	switch (*c++) {
 	    case L'{':
 		nest++;
