@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* expand.h: word expansion */
-/* (C) 2007-2010 magicant */
+/* (C) 2007-2016 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -24,14 +24,11 @@
 
 #define DEFAULT_IFS L" \t\n"
 
-/* characters that have special meanings in brace expansion and glob. */
-#define CHARS_ESCAPED L"\\{,}"
-
 /* characters that can be escaped with a backslash inside double-quotes. */
 #define CHARS_ESCAPABLE L"$`\"\\"
 
 /* type of tilde expansion */
-typedef enum { tt_none, tt_single, tt_multi, } tildetype_T;
+typedef enum { TT_NONE, TT_SINGLE, TT_MULTI, } tildetype_T;
 
 struct wordunit_T;
 struct plist_T;
@@ -43,19 +40,21 @@ extern _Bool expand_line(
 extern _Bool expand_multiple(
 	const struct wordunit_T *restrict w, struct plist_T *restrict list)
     __attribute__((nonnull(2)));
-extern wchar_t *expand_single(const struct wordunit_T *arg, tildetype_T tilde)
+extern wchar_t *expand_single(const struct wordunit_T *arg,
+	tildetype_T tilde, _Bool processquotes, _Bool escapeall)
+    __attribute__((malloc,warn_unused_result));
+extern wchar_t *expand_single_and_unescape(const struct wordunit_T *arg,
+	tildetype_T tilde, _Bool processquotes, _Bool escapeall)
     __attribute__((malloc,warn_unused_result));
 extern char *expand_single_with_glob(
 	const struct wordunit_T *arg, tildetype_T tilde)
     __attribute__((malloc,warn_unused_result));
-extern wchar_t *expand_string(const struct wordunit_T *w, _Bool esc)
-    __attribute__((malloc,warn_unused_result));
 
-extern wchar_t *split_next_field(
-	const wchar_t **sp, const wchar_t *ifs, _Bool noescape)
-    __attribute__((nonnull(1),malloc,warn_unused_result));
-extern void trim_trailing_ifsws(wchar_t *s, const wchar_t *ifs)
-    __attribute__((nonnull(1)));
+extern wchar_t *extract_fields(
+	const wchar_t *restrict s, const char *restrict split,
+	_Bool escaped, const wchar_t *restrict ifs,
+	struct plist_T *restrict dest)
+    __attribute__((nonnull));
 
 extern wchar_t *escape(const wchar_t *restrict s, const wchar_t *restrict t)
     __attribute__((nonnull(1),malloc,warn_unused_result));
