@@ -118,9 +118,11 @@ bool shopt_xtrace = false;
  * commands. */
 bool shopt_traceall = true;
 
+#if YASH_ENABLE_HISTORY
 /* If set, lines that start with a space are not saved in the history.
  * Corresponds to the --histspace option. */
 bool shopt_histspace = false;
+#endif
 /* Function definition commands are saved in the history only when this option
  * is set.
  * Corresponds to the --log option.
@@ -182,6 +184,9 @@ bool shopt_le_visiblebell = false;
 bool shopt_le_promptsp = true;
 /* If set, the right prompt is always visible on the screen. */
 bool shopt_le_alwaysrp = false;
+/* If set, auto-suggest the most probable command line as the user enters a
+ * command line. */
+bool shopt_le_predict = false;
 /* If set, debugging information is printed during command completion. */
 bool shopt_le_compdebug = false;
 #endif
@@ -214,7 +219,9 @@ static const struct option_T shell_options[] = {
     { 0,    0,    L"extendedglob",   &shopt_extendedglob,   true, },
     { 0,    L'f', L"glob",           &shopt_glob,           true, },
     { L'h', 0,    L"hashondef",      &shopt_hashondef,      true, },
+#if YASH_ENABLE_HISTORY
     { 0,    0,    L"histspace",      &shopt_histspace,      true, },
+#endif
     { 0,    0,    L"ignoreeof",      &shopt_ignoreeof,      true, },
     { L'i', 0,    L"interactive",    &is_interactive,       false, },
 #if YASH_ENABLE_LINEEDIT
@@ -222,6 +229,7 @@ static const struct option_T shell_options[] = {
     { 0,    0,    L"lecompdebug",    &shopt_le_compdebug,   true, },
     { 0,    0,    L"leconvmeta",     &shopt_le_yesconvmeta, true, },
     { 0,    0,    L"lenoconvmeta",   &shopt_le_noconvmeta,  true, },
+    { 0,    0,    L"lepredict",      &shopt_le_predict,     true, },
     { 0,    0,    L"lepromptsp",     &shopt_le_promptsp,    true, },
     { 0,    0,    L"levisiblebell",  &shopt_le_visiblebell, true, },
 #endif
@@ -314,7 +322,7 @@ static int set_normal_option(const struct xgetopt_T *opt, const wchar_t *arg,
 	struct shell_invocation_T *shell_invocation)
     __attribute__((nonnull(1)));
 #if YASH_ENABLE_TEST
-static struct option_T *find_option_unique(const wchar_t *s)
+static int test_option_unique(const wchar_t *s)
     __attribute__((nonnull,pure));
 #endif
 
@@ -404,7 +412,7 @@ int parse_short_option(void *const *argv, bool enable,
 		if (optname == NULL) {
 		    xerror(0, Ngt("the -%lc option requires an argument"),
 			    (wint_t) L'o');
-		    return special_builtin_syntax_error(Exit_ERROR);
+		    return special_builtin_error(Exit_ERROR);
 		}
 	    }
 
@@ -450,7 +458,7 @@ int parse_option_character(
     }
 
     xerror(0, Ngt("`%ls' is not a valid option"), (wchar_t[]) { opt, L'\0' });
-    return special_builtin_syntax_error(Exit_ERROR);
+    return special_builtin_error(Exit_ERROR);
 }
 
 /* Parses the long option at `xoptind' in the specified arguments and
@@ -569,7 +577,7 @@ int handle_search_result(plist_T *options, void *const *argv, bool enable,
 	case 0:
 	    pl_destroy(options);
 	    xerror(0, Ngt("`%ls' is not a valid option"), optstr);
-	    return special_builtin_syntax_error(Exit_ERROR);
+	    return special_builtin_error(Exit_ERROR);
 	case 1:
 	    if (noshelloptindex > 0) {
 		const struct option_T *opt = options->contents[0];
@@ -593,8 +601,7 @@ int handle_search_result(plist_T *options, void *const *argv, bool enable,
 				if (optarg == NULL) {
 				    xerror(0, Ngt("the --%ls option requires "
 						"an argument"), opt->longopt);
-				    return special_builtin_syntax_error(
-					    Exit_ERROR);
+				    return special_builtin_error(Exit_ERROR);
 				}
 				break;
 			    default:
@@ -607,7 +614,7 @@ int handle_search_result(plist_T *options, void *const *argv, bool enable,
 		    if (eq != NULL) {
 			xerror(0, Ngt("%ls: the --%ls option does not take "
 				    "an argument"), optstr, opt->longopt);
-			return special_builtin_syntax_error(Exit_ERROR);
+			return special_builtin_error(Exit_ERROR);
 		    }
 		    optarg = NULL;
 		}
@@ -628,7 +635,7 @@ int handle_search_result(plist_T *options, void *const *argv, bool enable,
 		    ((const struct xgetopt_T *) options->contents[i])->longopt);
 #endif
 	    pl_destroy(options);
-	    return special_builtin_syntax_error(Exit_ERROR);
+	    return special_builtin_error(Exit_ERROR);
     }
 }
 
@@ -641,7 +648,7 @@ int set_shell_option(const struct option_T *option, bool enable,
 	xerror(0, Ngt("the %ls option cannot be changed "
 		    "once the shell has been initialized"),
 		option->longopt);
-	return special_builtin_syntax_error(Exit_ERROR);
+	return special_builtin_error(Exit_ERROR);
     }
 
     *option->optp = enable;
@@ -683,7 +690,8 @@ int set_shell_option(const struct option_T *option, bool enable,
     }
 #endif
 
-    return (*option->optp == enable) ? Exit_SUCCESS : Exit_FAILURE;
+    return (*option->optp == enable) ?
+	    Exit_SUCCESS : special_builtin_error(Exit_FAILURE);
 }
 
 #if YASH_ENABLE_LINEEDIT
@@ -783,31 +791,37 @@ void set_lineedit_option(enum shopt_lineedit_T v)
  * as the argument to the "-o" option. */
 bool is_valid_option_name(const wchar_t *s)
 {
-    return find_option_unique(s) != NULL;
+    return test_option_unique(s);
 }
 
 /* Returns true iff the specified string is a valid option name that can be used
  * as the argument to the "-o" option and the option is enabled. */
 bool option_is_enabled(const wchar_t *s)
 {
-    struct option_T *opt = find_option_unique(s);
-    return opt != NULL && *opt->optp;
+    return test_option_unique(s) == 2;
 }
 
-/* If the specified string is a valid option name that can be used as the
- * argument to the "-o" option, then returns the option. Otherwise, NULL is
- * returned. */
-struct option_T *find_option_unique(const wchar_t *s)
+/* Tests if the specified string is a valid option name that can be used
+ * as the argument to the "-o" option. Returns:
+ * 0 if the string is not a valid option name,
+ * 1 if it is a valid option that is currently disabled, or
+ * 2 if it is a valid option that is currently enabled. */
+int test_option_unique(const wchar_t *s)
 {
     plist_T options;
     pl_init(&options);
 
-    collect_matching_shell_options(s, &options);
+    size_t nooptindex = collect_matching_shell_options(s, &options);
+    int result = (options.length == 1);
 
-    struct option_T *opt = (options.length == 1) ? options.contents[0] : NULL;
+    if (result) {
+	struct option_T *opt = options.contents[0];
+	bool negated = (nooptindex == 0);
+	result += *opt->optp ^ negated;
+    }
 
     pl_destroy(&options);
-    return opt;
+    return result;
 }
 
 #endif /* YASH_ENABLE_TEST */
@@ -903,7 +917,7 @@ int set_builtin_print_current_settings(void)
 
     for (const struct option_T *o = shell_options; o->longopt != NULL; o++) {
 	if (!xprintf("%-15ls %s\n", o->longopt, vals[(bool) *o->optp]))
-	    return Exit_FAILURE;
+	    return special_builtin_error(Exit_FAILURE);
     }
 
     return Exit_SUCCESS;
@@ -914,7 +928,7 @@ int set_builtin_print_restoring_commands(void)
     for (const struct option_T *o = shell_options; o->longopt != NULL; o++) {
 	if (o->resettable)
 	    if (!xprintf("set %co %ls\n", *o->optp ? '-' : '+', o->longopt))
-		return Exit_FAILURE;
+		return special_builtin_error(Exit_FAILURE);
     }
 
     return Exit_SUCCESS;
