@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* parser.c: syntax parser */
-/* (C) 2007-2016 magicant */
+/* (C) 2007-2017 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -248,6 +248,8 @@ static bool is_name_by_predicate(const wchar_t *s, bool predicate(wchar_t))
     __attribute__((pure,nonnull));
 static bool is_portable_name(const wchar_t *s)
     __attribute__((pure,nonnull));
+static bool is_literal_function_name(const wordunit_T *wu)
+    __attribute__((pure));
 
 
 /* Checks if the specified character can be used in a portable variable name.
@@ -356,6 +358,22 @@ bool is_keyword(const wchar_t *s)
     }
 }
 
+bool is_literal_function_name(const wordunit_T *wu)
+{
+    if (wu == NULL)
+	return false;
+    if (wu->next != NULL)
+	return false;
+    if (wu->wu_type != WT_STRING)
+	return false;
+
+    const wchar_t *s = wu->wu_string;
+    if (iswdigit(s[0]))
+	return false;
+
+    return (posixly_correct ? is_portable_name : is_name)(s);
+}
+
 
 /********** Parser **********/
 
@@ -381,8 +399,6 @@ typedef struct parsestate_T {
  * aliases: a list of alias substitutions that were performed at the current
  *         position. */
 
-typedef enum { AT_NONE, AT_GLOBAL, AT_ALL, } aliastype_T;
-
 static void serror(parsestate_T *restrict ps, const char *restrict format, ...)
     __attribute__((nonnull(1,2),format(printf,2,3)));
 static inputresult_T read_more_input(parsestate_T *ps)
@@ -413,6 +429,10 @@ static const wchar_t *check_opening_token(parsestate_T *ps)
     __attribute__((nonnull));
 static const wchar_t *check_closing_token(parsestate_T *ps)
     __attribute__((nonnull));
+static bool psubstitute_alias(parsestate_T *ps, substaliasflags_T f)
+    __attribute__((nonnull));
+static void psubstitute_alias_recursive(parsestate_T *ps, substaliasflags_T f)
+    __attribute__((nonnull));
 static and_or_T *parse_command_list(parsestate_T *ps, bool toeol)
     __attribute__((nonnull,malloc,warn_unused_result));
 static and_or_T *parse_compound_list(parsestate_T *ps)
@@ -440,7 +460,7 @@ static void **parse_words_to_paren(parsestate_T *ps)
     __attribute__((nonnull,malloc,warn_unused_result));
 static redir_T *tryparse_redirect(parsestate_T *ps)
     __attribute__((nonnull,malloc,warn_unused_result));
-static wordunit_T *parse_word(parsestate_T *ps, aliastype_T type)
+static wordunit_T *parse_word(parsestate_T *ps, bool globalaliases)
     __attribute__((nonnull,malloc,warn_unused_result));
 static wordunit_T *parse_word_to(parsestate_T *ps, bool testfunc(wchar_t c))
     __attribute__((nonnull,malloc,warn_unused_result));
@@ -483,8 +503,8 @@ static void **parse_case_patterns(parsestate_T *ps)
     __attribute__((nonnull,malloc,warn_unused_result));
 static command_T *parse_function(parsestate_T *ps)
     __attribute__((nonnull,malloc,warn_unused_result));
-static command_T *tryparse_function(parsestate_T *ps)
-    __attribute__((nonnull,malloc,warn_unused_result));
+static command_T *try_reparse_as_function(parsestate_T *ps, command_T *c)
+    __attribute__((nonnull,warn_unused_result));
 static void read_heredoc_contents(parsestate_T *ps, redir_T *redir)
     __attribute__((nonnull));
 static void read_heredoc_contents_without_expansion(
@@ -638,6 +658,7 @@ void line_continuation(parsestate_T *ps, size_t index)
     assert(ps->src.contents[index] == L'\\' &&
 	    ps->src.contents[index + 1] == L'\n');
     wb_remove(&ps->src, index, 2);
+    shift_aliaslist_index(ps->aliases, index + 1, -2);
     ps->info->lineno++;
     if (ps->src.contents[index] == L'\0')
 	read_more_input(ps);
@@ -848,6 +869,23 @@ const wchar_t *check_closing_token(parsestate_T *ps)
     return NULL;
 }
 
+/* Performs alias substitution with the given parse state. */
+bool psubstitute_alias(parsestate_T *ps, substaliasflags_T flags)
+{
+    size_t len = count_name_length(ps, is_alias_name_char);
+    return substitute_alias_range(
+	    &ps->src, ps->index, ps->index + len, &ps->aliases, flags);
+}
+
+/* Performs alias substitution recursively. This should not be used where the
+ * substitution result may be recognized as a keyword, since keywords should not
+ * be alias-substituted. */
+void psubstitute_alias_recursive(parsestate_T *ps, substaliasflags_T flags)
+{
+    while (psubstitute_alias(ps, flags))
+	skip_blanks_and_comment(ps);
+}
+
 /* Parses commands.
  * If `toeol' is true, commands are parsed up to the end of the current input;
  * otherwise, up to the next closing token.
@@ -1087,32 +1125,21 @@ command_T *parse_command(parsestate_T *ps)
     if (t != NULL)
 	return parse_compound_command(ps, t);
 
-    if (ps->enable_alias && count_name_length(ps, is_alias_name_char) > 0) {
-	substaliasflags_T flags = AF_NONGLOBAL | AF_NORECUR;
-	if (substitute_alias(&ps->src, ps->index, &ps->aliases, flags)) {
-	    ps->reparse = true;
-	    return NULL;
-	}
+    if (ps->enable_alias && psubstitute_alias(ps, AF_NONGLOBAL)) {
+	ps->reparse = true;
+	return NULL;
     }
 
-    command_T *result = tryparse_function(ps);
-    if (result != NULL)
-	return result;
-
     /* parse as a simple command */
-    redir_T **redirlastp;
-    result = xmalloc(sizeof *result);
+    command_T *result = xmalloc(sizeof *result);
     result->next = NULL;
     result->refcount = 1;
     result->c_lineno = ps->info->lineno;
     result->c_type = CT_SIMPLE;
-    redirlastp = parse_assignments_and_redirects(ps, result);
+    redir_T **redirlastp = parse_assignments_and_redirects(ps, result);
     result->c_words = parse_words_and_redirects(ps, redirlastp, true);
 
-    ensure_buffer(ps, 1);
-    if (ps->src.contents[ps->index] == L'(')
-	serror(ps, Ngt("invalid use of `%lc'"), (wint_t) L'(');
-    return result;
+    return try_reparse_as_function(ps, result);
 }
 
 /* Parses assignments and redirections.
@@ -1142,11 +1169,8 @@ redir_T **parse_assignments_and_redirects(parsestate_T *ps, command_T *c)
 	} else {
 	    break;
 	}
-	if (ps->enable_alias && count_name_length(ps, is_alias_name_char) > 0) {
-	    substitute_alias(
-		    &ps->src, ps->index, &ps->aliases, AF_NONGLOBAL);
-	    skip_blanks_and_comment(ps);
-	}
+	if (ps->enable_alias)
+	    psubstitute_alias_recursive(ps, AF_NONGLOBAL);
     }
     return redirlastp;
 }
@@ -1169,16 +1193,12 @@ void **parse_words_and_redirects(
     pl_init(&wordlist);
     while (ensure_buffer(ps, 1),
 	    !is_command_delimiter_char(ps->src.contents[ps->index])) {
-	if (!first && ps->enable_alias) {
-	    if (count_name_length(ps, is_alias_name_char) > 0) {
-		substitute_alias(&ps->src, ps->index, &ps->aliases, 0);
-		skip_blanks_and_comment(ps);
-	    }
-	}
+	if (!first && ps->enable_alias)
+	    psubstitute_alias_recursive(ps, 0);
 	if ((redir = tryparse_redirect(ps)) != NULL) {
 	    *redirlastp = redir;
 	    redirlastp = &redir->next;
-	} else if ((word = parse_word(ps, AT_NONE)) != NULL) {
+	} else if ((word = parse_word(ps, false)) != NULL) {
 	    pl_add(&wordlist, word);
 	    skip_blanks_and_comment(ps);
 	    first = false;
@@ -1195,9 +1215,8 @@ void **parse_words_and_redirects(
 void parse_redirect_list(parsestate_T *ps, redir_T **lastp)
 {
     for (;;) {
-	if (!posixly_correct && ps->enable_alias)
-	    if (count_name_length(ps, is_alias_name_char) > 0)
-		substitute_alias(&ps->src, ps->index, &ps->aliases, 0);
+	if (ps->enable_alias)
+	    psubstitute_alias_recursive(ps, 0);
 
 	redir_T *redir = tryparse_redirect(ps);
 	if (redir == NULL)
@@ -1226,7 +1245,7 @@ assign_T *tryparse_assignment(parsestate_T *ps)
     ensure_buffer(ps, 1);
     if (posixly_correct || ps->src.contents[ps->index] != L'(') {
 	result->a_type = A_SCALAR;
-	result->a_scalar = parse_word(ps, AT_NONE);
+	result->a_scalar = parse_word(ps, false);
     } else {
 	ps->index++;
 	skip_to_next_token(ps);
@@ -1250,7 +1269,7 @@ void **parse_words_to_paren(parsestate_T *ps)
 
     pl_init(&list);
     while (ps->src.contents[ps->index] != L')') {
-	wordunit_T *word = parse_word(ps, AT_GLOBAL);
+	wordunit_T *word = parse_word(ps, true);
 	if (word != NULL)
 	    pl_add(&list, word);
 	else
@@ -1357,7 +1376,7 @@ reparse:
     }
     skip_blanks_and_comment(ps);
     if (result->rd_type != RT_HERE && result->rd_type != RT_HERERT) {
-	result->rd_filename = parse_word(ps, AT_GLOBAL);
+	result->rd_filename = parse_word(ps, true);
 	if (result->rd_filename == NULL) {
 	    serror(ps, Ngt("the redirection target is missing"));
 	    free(result);
@@ -1390,25 +1409,12 @@ parse_command:
     return result;
 }
 
-/* Expands an alias, if any, and parses a word at the current position.
- * `type' specifies the type of aliases to be expanded.  */
-wordunit_T *parse_word(parsestate_T *ps, aliastype_T type)
+/* Parses a word at the current position. If `globalaliases' is true, global
+ * aliases are substituted before the word is parsed. */
+wordunit_T *parse_word(parsestate_T *ps, bool globalaliases)
 {
-    if (ps->enable_alias) {
-	switch (type) {
-	case AT_NONE:
-	    break;
-	case AT_GLOBAL:
-	case AT_ALL:
-	    if (count_name_length(ps, is_alias_name_char) > 0) {
-		substaliasflags_T flags =
-		    (type == AT_GLOBAL) ? 0 : AF_NONGLOBAL;
-		substitute_alias(&ps->src, ps->index, &ps->aliases, flags);
-	    }
-	    skip_blanks_and_comment(ps);
-	    break;
-	}
-    }
+    if (globalaliases && ps->enable_alias)
+	psubstitute_alias_recursive(ps, 0);
 
     return parse_word_to(ps, is_token_delimiter_char);
 }
@@ -1982,7 +1988,7 @@ fail:
 wchar_t *parse_word_as_wcs(parsestate_T *ps)
 {
     size_t startindex = ps->index;
-    wordfree(parse_word(ps, AT_GLOBAL));
+    wordfree(parse_word(ps, true));
     assert(startindex <= ps->index);
     return xwcsndup(&ps->src.contents[startindex], ps->index - startindex);
 }
@@ -2156,6 +2162,7 @@ command_T *parse_for(parsestate_T *ps)
     }
     result->c_forname = name;
 
+parse_in:;
     bool on_next_line = skip_to_next_token(ps);
     ensure_buffer(ps, 3);
     if (has_token(ps, L"in")) {
@@ -2169,6 +2176,8 @@ command_T *parse_for(parsestate_T *ps)
 	}
 	if (ps->src.contents[ps->index] == L';')
 	    ps->index++;
+    } else if (psubstitute_alias(ps, 0)) {
+	goto parse_in;
     } else {
 	result->c_forwords = NULL;
 	if (ps->src.contents[ps->index] == L';') {
@@ -2177,17 +2186,23 @@ command_T *parse_for(parsestate_T *ps)
 		serror(ps, Ngt("`;' cannot appear on a new line"));
 	}
     }
+
+parse_do:
     skip_to_next_token(ps);
     ensure_buffer(ps, 3);
     if (has_token(ps, L"do"))
 	ps->index += 2;
+    else if (psubstitute_alias(ps, 0))
+	goto parse_do;
     else
 	serror(ps, Ngt("`%ls' is missing"), L"do");
 	// print_errmsg_token_missing(ps, L"do");
+
     result->c_forcmds = parse_compound_list(ps);
     if (posixly_correct && result->c_forcmds == NULL)
 	serror(ps, Ngt("commands are missing between `%ls' and `%ls'"),
 		L"do", L"done");
+
     ensure_buffer(ps, 5);
     if (has_token(ps, L"done"))
 	ps->index += 4;
@@ -2245,19 +2260,24 @@ command_T *parse_case(parsestate_T *ps)
     result->c_type = CT_CASE;
     result->c_lineno = ps->info->lineno;
     result->c_redirs = NULL;
-    result->c_casword = parse_word(ps, AT_GLOBAL);
+    result->c_casword = parse_word(ps, true);
     if (result->c_casword == NULL)
 	serror(ps, Ngt("a word is required after `%ls'"), L"case");
+
+parse_in:
     skip_to_next_token(ps);
     ensure_buffer(ps, 3);
     if (has_token(ps, L"in")) {
 	ps->index += 2;
 	result->c_casitems = parse_case_list(ps);
+    } else if (psubstitute_alias(ps, 0)) {
+	goto parse_in;
     } else {
 	serror(ps, Ngt("`%ls' is missing"), L"in");
 	// print_errmsg_token_missing(ps, L"in");
 	result->c_casitems = NULL;
     }
+
     ensure_buffer(ps, 5);
     if (has_token(ps, L"esac"))
 	ps->index += 4;
@@ -2277,6 +2297,8 @@ caseitem_T *parse_case_list(parsestate_T *ps)
 	ensure_buffer(ps, 5);
 	if (has_token(ps, L"esac"))
 	    break;
+	if (psubstitute_alias(ps, 0))
+	    continue;
 
 	caseitem_T *ci = xmalloc(sizeof *ci);
 	*lastp = ci;
@@ -2331,8 +2353,9 @@ void **parse_case_patterns(parsestate_T *ps)
 	    }
 	    break;
 	}
-	pl_add(&wordlist, parse_word(ps, AT_GLOBAL));
+	pl_add(&wordlist, parse_word(ps, true));
 	skip_blanks_and_comment(ps);
+	psubstitute_alias_recursive(ps, 0);
 	ensure_buffer(ps, 1);
 	if (ps->src.contents[ps->index] == L'|') {
 	    predecessor = L"|";
@@ -2365,89 +2388,100 @@ command_T *parse_function(parsestate_T *ps)
     result->c_type = CT_FUNCDEF;
     result->c_lineno = ps->info->lineno;
     result->c_redirs = NULL;
-    result->c_funcname = parse_word(ps, AT_GLOBAL);
+    result->c_funcname = parse_word(ps, true);
     if (result->c_funcname == NULL)
 	serror(ps, Ngt("a word is required after `%ls'"), L"function");
     skip_blanks_and_comment(ps);
 
-    /* parse parentheses */
+    bool paren = false;
+parse_parentheses:;
     size_t saveindex = ps->index;
     if (ps->src.contents[ps->index] == L'(') {
 	ps->index++;
+parse_close_parenthesis:
 	skip_blanks_and_comment(ps);
 	if (ps->src.contents[ps->index] == L')')
-	    ps->index++;
+	    paren = true, ps->index++;
+	else if (psubstitute_alias(ps, AF_NONGLOBAL))
+	    goto parse_close_parenthesis;
 	else
 	    ps->index = saveindex;
     }
     skip_to_next_token(ps);
 
-    /* parse function body */
+parse_function_body:;
     const wchar_t *t = check_opening_token(ps);
-    if (t == NULL) {
+    if (t != NULL) {
+	result->c_funcbody = parse_compound_command(ps, t);
+    } else if (psubstitute_alias(ps, 0)) {
+	if (paren)
+	    goto parse_function_body;
+	else
+	    goto parse_parentheses;
+    } else {
 	serror(ps, Ngt("a function body must be a compound command"));
 	result->c_funcbody = NULL;
-    } else {
-	result->c_funcbody = parse_compound_command(ps, t);
     }
 
     return result;
 }
 
-/* Parses a function definition if any.
- * This function may process line continuations, which increase the line
- * number. */
-command_T *tryparse_function(parsestate_T *ps)
+/* Parses (part of) a function definition command that does not start with the
+ * "function" keyword. This function must be called just after a simple command
+ * has been parsed, which is given as `c'. If the next character is '(', it
+ * should signify a function definition, so this function continues parsing the
+ * rest of it. Otherwise, `c' is returned intact.
+ * If successful, `c' is directly modified to the function definition parsed. */
+command_T *try_reparse_as_function(parsestate_T *ps, command_T *c)
 {
-    size_t saveindex = ps->index;
-    unsigned long lineno = ps->info->lineno;
+    // ensure_buffer(ps, 1);
+    if (ps->src.contents[ps->index] != L'(') // not a function definition?
+	return c;
 
-    if (iswdigit(ps->src.contents[ps->index]))
-	goto fail;
+    /* If this is a function definition, there must be exactly one command word
+     * before '('. */
+    assert(c->c_type == CT_SIMPLE);
+    if (c->c_redirs != NULL || c->c_assigns != NULL
+	    || c->c_words[0] == NULL || c->c_words[1] != NULL) {
+	serror(ps, Ngt("invalid use of `%lc'"), (wint_t) L'(');
+	return c;
+    }
 
-    size_t namelen = count_name_length(ps,
-	    posixly_correct ? is_portable_name_char : is_name_char);
-    ps->index += namelen;
-    if (namelen == 0 || !is_token_delimiter_char(ps->src.contents[ps->index]))
-	goto fail;
-    skip_blanks_and_comment(ps);
+    /* The name must be valid. */
+    wordunit_T *name = c->c_words[0];
+    if (!is_literal_function_name(name)) {
+	serror(ps, Ngt("invalid function name"));
+	return c;
+    }
 
-    /* parse parentheses */
-    if (ps->src.contents[ps->index] != L'(')
-	goto fail;
+    /* Skip '('. */
     ps->index++;
     skip_blanks_and_comment(ps);
+
+    /* Parse ')'. */
+    psubstitute_alias_recursive(ps, 0);
+    // ensure_buffer(ps, 1);
     if (ps->src.contents[ps->index] != L')') {
 	serror(ps, Ngt("`(' must be followed by `)' in a function definition"));
-	return NULL;
+	return c;
     }
     ps->index++;
+parse_function_body:
     skip_to_next_token(ps);
 
-    /* parse function body */
     const wchar_t *t = check_opening_token(ps);
     if (t == NULL) {
+	if (psubstitute_alias(ps, 0))
+	    goto parse_function_body;
 	serror(ps, Ngt("a function body must be a compound command"));
-	return NULL;
+	return c;
     }
-    command_T *body = parse_compound_command(ps, t);
-    command_T *result = xmalloc(sizeof *result);
-    result->next = NULL;
-    result->refcount = 1;
-    result->c_type = CT_FUNCDEF;
-    result->c_lineno = lineno;
-    result->c_redirs = NULL;
-    result->c_funcname = xmalloc(sizeof *result->c_funcname);
-    result->c_funcname->next = NULL;
-    result->c_funcname->wu_type = WT_STRING;
-    result->c_funcname->wu_string =
-	xwcsndup(&ps->src.contents[saveindex], namelen);
-    result->c_funcbody = body;
-    return result;
+    free(c->c_words);
 
-fail:
-    ps->index = saveindex;
-    return NULL;
+    c->c_type = CT_FUNCDEF;
+    c->c_funcname = name;
+    c->c_funcbody = parse_compound_command(ps, t);
+    return c;
 }
 
 /* Reads the contents of a here-document. */
@@ -3111,9 +3145,7 @@ void print_function_definition(
 {
     assert(c->c_type == CT_FUNCDEF);
 
-    if (c->c_funcname->next != NULL
-	    || c->c_funcname->wu_type != WT_STRING
-	    || !is_name(c->c_funcname->wu_string))
+    if (!is_literal_function_name(c->c_funcname))
 	wb_cat(&pr->buffer, L"function ");
     print_word(pr, c->c_funcname, indent);
     wb_cat(&pr->buffer, L"()");
