@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* redir.c: manages file descriptors and provides functions for redirections */
-/* (C) 2007-2016 magicant */
+/* (C) 2007-2017 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -334,7 +334,7 @@ bool open_redirections(const redir_T *r, savefd_T **save)
 	}
 
 	/* expand rd_filename */
-	char *INIT(filename);
+	char *INIT(filename, "");
 	switch (r->rd_type) {
 	    case RT_INPUT:  case RT_OUTPUT:  case RT_CLOBBER:  case RT_APPEND:
 	    case RT_INOUT:  case RT_DUPIN:   case RT_DUPOUT:   case RT_PIPE:
@@ -362,6 +362,7 @@ bool open_redirections(const redir_T *r, savefd_T **save)
 	    if (!shopt_clobber && !is_irregular_file(filename)) {
 		flags = O_WRONLY | O_CREAT | O_EXCL;
 	    } else {
+		/* falls thru! */
 	case RT_CLOBBER:
 		flags = O_WRONLY | O_CREAT | O_TRUNC;
 	    }
@@ -894,15 +895,12 @@ void clear_savefd(savefd_T *save)
 }
 
 /* Redirects the standard input to "/dev/null" if job control is off and the
- * standard input is not yet redirected.
- * If `posixly_correct' is true, the condition is slightly different:
- * "if non-interactive" rather than "if job control is off". */
+ * standard input is not yet redirected. */
 void maybe_redirect_stdin_to_devnull(void)
 {
     int fd;
 
-    if ((posixly_correct ? is_interactive : do_job_control)
-	    || is_stdin_redirected)
+    if (do_job_control || is_stdin_redirected)
 	return;
 
     if (xclose(STDIN_FILENO) < 0)
