@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* sig.c: signal handling */
-/* (C) 2007-2016 magicant */
+/* (C) 2007-2017 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -654,22 +654,18 @@ enum wait_for_input_T wait_for_input(int fd, bool trap, int timeout)
 /* Handles SIGCHLD if caught. */
 void handle_sigchld(void)
 {
-    static bool print_status = false;
-
     if (sigchld_received) {
         sigchld_received = false;
         do_wait();
-	print_status = true;
     }
 
-    if (print_status) {
+    if (any_job_status_has_changed()) {
 	/* print job status if the notify option is set */
 #if YASH_ENABLE_LINEEDIT
 	if (le_state & LE_STATE_ACTIVE) {
 	    if (!(le_state & LE_STATE_COMPLETING)) {
 		if (shopt_notify || shopt_notifyle) {
 		    le_suspend_readline();
-		    print_status = false;
 		    print_job_status_all();
 		    le_resume_readline();
 		}
@@ -682,7 +678,6 @@ void handle_sigchld(void)
 	    sigaddset(&ss, SIGTTOU);
 	    sigemptyset(&savess);
 	    sigprocmask(SIG_BLOCK, &ss, &savess);
-	    print_status = false;
 	    print_job_status_all();
 	    sigprocmask(SIG_SETMASK, &savess, NULL);
 	}
@@ -728,10 +723,12 @@ int handle_traps(void)
 		    le_suspend_readline();
 #endif
 		    struct execstate_T *execstate = save_execstate();
+		    reset_execstate(true);
 		    signum = handled_signal = s->no;
 		    command = xwcsdup(command);
 		    exec_wcs(command, "trap", false);
 		    free(command);
+		    cancel_return();
 		    restore_execstate(execstate);
 		    laststatus = savelaststatus;
 		}
@@ -750,10 +747,12 @@ int handle_traps(void)
 		    le_suspend_readline();
 #endif
 		    struct execstate_T *execstate = save_execstate();
+		    reset_execstate(true);
 		    signum = handled_signal = sigrtmin + i;
 		    command = xwcsdup(command);
 		    exec_wcs(command, "trap", false);
 		    free(command);
+		    cancel_return();
 		    restore_execstate(execstate);
 		    laststatus = savelaststatus;
 		}
@@ -790,7 +789,7 @@ void execute_exit_trap(void)
 #endif
 	savelaststatus = laststatus;
 	command = xwcsdup(command);
-	reset_execstate();
+	reset_execstate(true);
 	exec_wcs(command, "EXIT trap", false);
 	free(command);
 	savelaststatus = -1;
@@ -1153,7 +1152,7 @@ int trap_builtin(int argc, void **argv)
 		return print_builtin_help(ARGV(0));
 #endif
 	    default:
-		return special_builtin_syntax_error(Exit_ERROR);
+		return special_builtin_error(Exit_ERROR);
 	}
     }
 
@@ -1167,7 +1166,7 @@ int trap_builtin(int argc, void **argv)
 	    if (!sigismember(&printed, s->no)) {
 		sigaddset(&printed, s->no);
 		if (!print_trap(s->name, trap_command[sigindex(s->no)]))
-		    return Exit_FAILURE;
+		    return special_builtin_error(Exit_FAILURE);
 	    }
 	}
 #if defined SIGRTMIN && defined SIGRTMAX
@@ -1176,7 +1175,7 @@ int trap_builtin(int argc, void **argv)
 	    if (sigrtmin + i > sigrtmax)
 		break;
 	    if (!print_trap(get_signal_name(sigrtmin + i), rttrap_command[i]))
-		return Exit_FAILURE;
+		return special_builtin_error(Exit_FAILURE);
 	}
 #endif
     } else if (print) {
@@ -1197,12 +1196,12 @@ int trap_builtin(int argc, void **argv)
 		int index = signum - sigrtmin;
 		if (index < RTSIZE)
 		    if (!print_trap(name, rttrap_command[index]))
-			return Exit_FAILURE;
+			return special_builtin_error(Exit_FAILURE);
 	    } else
 #endif
 	    {
 		if (!print_trap(name, trap_command[sigindex(signum)]))
-		    return Exit_FAILURE;
+		    return special_builtin_error(Exit_FAILURE);
 	    }
 	} while (++xoptind < argc);
     } else {
@@ -1217,8 +1216,7 @@ int trap_builtin(int argc, void **argv)
 	} else {
 	    command = ARGV(xoptind++);
 	    if (xoptind == argc)
-		return special_builtin_syntax_error(
-			insufficient_operands_error(2));
+		return special_builtin_error(insufficient_operands_error(2));
 
 	    if (wcscmp(command, L"-") == 0)
 		command = NULL;
