@@ -1,6 +1,6 @@
 /* Yash: yet another shell */
 /* yash.c: basic functions of the shell */
-/* (C) 2007-2013 magicant */
+/* (C) 2007-2016 magicant */
 
 /* This program is free software: you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -31,9 +31,7 @@
 #include <string.h>
 #include <unistd.h>
 #include <wchar.h>
-#if YASH_ENABLE_ALIAS
-# include "alias.h"
-#endif
+#include "alias.h"
 #include "builtin.h"
 #include "configm.h"
 #include "exec.h"
@@ -59,7 +57,10 @@ static struct input_file_info_T *new_input_file_info(int fd, size_t bufsize)
     __attribute__((malloc,warn_unused_result));
 static void execute_profile(const wchar_t *profile);
 static void execute_rcfile(const wchar_t *rcfile);
-static void execute_file(const wchar_t *path);
+static bool execute_file_in_home(const wchar_t *path)
+    __attribute__((nonnull));
+static bool execute_file(const wchar_t *path);
+static bool execute_file_mbs(const char *path);
 static void print_help(void);
 static void print_version(void);
 
@@ -145,9 +146,7 @@ int main(int argc, char **argv)
     init_shellfds();
     init_job();
     init_builtin();
-#if YASH_ENABLE_ALIAS
     init_alias();
-#endif
 
     struct shell_invocation_T options = {
 	.profile = NULL, .rcfile = NULL,
@@ -260,20 +259,16 @@ struct input_file_info_T *new_input_file_info(int fd, size_t bufsize)
 /* Executes "$HOME/.yash_profile". */
 void execute_profile(const wchar_t *profile)
 {
-    if (profile != NULL) {
+    if (profile != NULL)
 	execute_file(profile);
-	return;
-    }
-
-    wchar_t *path =
-	parse_and_expand_string(L"$HOME/.yash_profile", NULL, false);
-    execute_file(path);
-    free(path);
+    else
+	execute_file_in_home(L".yash_profile");
 }
 
 /* Executes the initialization file.
  * `rcfile' is the filename to source.
- * If `rcfile' is NULL, it defaults to "$HOME/.yashrc".
+ * If `rcfile' is NULL, it defaults to "$HOME/.yashrc" and if the file cannot be
+ * read it falls back to "initialization/default" from $YASH_LOADPATH.
  * In the POSIXly-correct mode, `rcfile' is ignored and "$ENV" is used. */
 void execute_rcfile(const wchar_t *rcfile)
 {
@@ -291,30 +286,65 @@ void execute_rcfile(const wchar_t *rcfile)
 	execute_file(rcfile);
 	return;
     }
+    if (execute_file_in_home(L".yashrc"))
+	return;
 
-    wchar_t *path = parse_and_expand_string(L"$HOME/.yashrc", NULL, false);
-    execute_file(path);
+    char *path = which("initialization/default", get_path_array(PA_LOADPATH),
+	    is_readable_regular);
+    execute_file_mbs(path);
     free(path);
 }
 
-/* Executes the specified file if `path' is non-NULL. */
-void execute_file(const wchar_t *path)
+/* Executes the specified file. The `path' must be relative to $HOME. */
+bool execute_file_in_home(const wchar_t *path)
+{
+    const wchar_t *home = getvar(L VAR_HOME);
+    if (home == NULL || home[0] == L'\0')
+	return false;
+
+    xwcsbuf_T fullpath;
+    wb_init(&fullpath);
+    wb_cat(&fullpath, home);
+    if (fullpath.contents[fullpath.length - 1] != L'/')
+	wb_wccat(&fullpath, L'/');
+    wb_cat(&fullpath, path);
+
+    bool executed = execute_file(fullpath.contents);
+
+    wb_destroy(&fullpath);
+
+    return executed;
+}
+
+/* Executes the specified file if `path' is non-NULL.
+ * Returns true iff the file was found. */
+bool execute_file(const wchar_t *path)
 {
     if (path == NULL)
-	return;
+	return false;
 
     char *mbspath = malloc_wcstombs(path);
-    if (mbspath == NULL)
-	return;
-
-    int fd = move_to_shellfd(open(mbspath, O_RDONLY));
-    if (fd >= 0) {
-	exec_input(fd, mbspath, XIO_SUBST_ALIAS);
-	remove_shellfd(fd);
-	xclose(fd);
-    }
-
+    bool executed = execute_file_mbs(mbspath);
     free(mbspath);
+    return executed;
+}
+
+/* Executes the specified file if `path' is non-NULL.
+ * Returns true iff the file was found. */
+bool execute_file_mbs(const char *path)
+{
+    if (path == NULL)
+	return false;
+
+    int fd = move_to_shellfd(open(path, O_RDONLY));
+    if (fd < 0)
+	return false;
+
+    exec_input(fd, path, XIO_SUBST_ALIAS);
+    cancel_return();
+    remove_shellfd(fd);
+    xclose(fd);
+    return true;
 }
 
 /* Exits the shell with the specified exit status.
@@ -378,9 +408,6 @@ void print_version(void)
 #ifndef NDEBUG
 		" * DEBUG\n"
 #endif
-#if YASH_ENABLE_ALIAS
-		" * alias\n"
-#endif
 #if YASH_ENABLE_ARRAY
 		" * array\n"
 #endif
@@ -429,9 +456,7 @@ void exec_wcs(const wchar_t *code, const char *name, bool finally_exit)
     struct parseparam_T pinfo = {
 	.print_errmsg = true,
 	.enable_verbose = false,
-#if YASH_ENABLE_ALIAS
 	.enable_alias = true,
-#endif
 	.filename = name,
 	.lineno = 1,
 	.input = input_wcs,
@@ -453,9 +478,7 @@ void exec_input(int fd, const char *name, exec_input_options_T options)
     struct parseparam_T pinfo = {
 	.print_errmsg = true,
 	.enable_verbose = true,
-#if YASH_ENABLE_ALIAS
 	.enable_alias = options & XIO_SUBST_ALIAS,
-#endif
 	.filename = name,
 	.lineno = 1,
 	.interactive = options & XIO_INTERACTIVE,
@@ -542,7 +565,6 @@ void parse_and_exec(parseparam_T *pinfo, bool finally_exit)
 	}
     }
 out:
-    cancel_return();
     if (finally_exit)
 	exit_shell();
 }
@@ -585,12 +607,12 @@ int exit_builtin(int argc, void **argv)
 		return print_builtin_help(ARGV(0));
 #endif
 	    default:
-		return special_builtin_syntax_error(Exit_ERROR);
+		return special_builtin_error(Exit_ERROR);
 	}
     }
 
     if (!validate_operand_count(argc - xoptind, 0, 1))
-	return special_builtin_syntax_error(Exit_ERROR);
+	return special_builtin_error(Exit_ERROR);
 
     size_t sjc;
     if (is_interactive_now && !forceexit && (sjc = stopped_job_count()) > 0) {
@@ -610,7 +632,7 @@ int exit_builtin(int argc, void **argv)
 	if (!xwcstoi(statusstr, 10, &status) || status < 0) {
 	    xerror(0, Ngt("`%ls' is not a valid integer"), statusstr);
 	    status = Exit_ERROR;
-	    special_builtin_syntax_error(status);
+	    special_builtin_error(status);
 	}
     } else {
 	status = -1;
